'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.TransportEngines import BaseEngine
from CiscoAutomationFramework.IOS import IOS
from CiscoAutomationFramework.IOSXE import IOSXE
from CiscoAutomationFramework.NXOS import NXOS
from inspect import getmodule
from time import sleep


def detect_firmware(transport):
    if not isinstance(transport, BaseEngine):
        raise TypeError(f'transport argument MUST be an instance of {getmodule(BaseEngine).__name__}.{BaseEngine.__name__}')

    _ = transport.send_command_get_output('terminal length 0')
    show_version = transport.send_command_get_output('show version')

    results = {'IOSXE': 0, 'IOS': 0, 'NXOS': 0, 'ASA': 0}
    for line in show_version[:10]:
        if 'ios-xe' in line.lower() or 'ios xe' in line.lower():
            results['IOSXE'] += 1
        elif 'ios' in line.lower():
            results['IOS'] += 1
        elif 'nx-os' in line.lower():
            results['NXOS'] += 1
        elif 'adaptive security appliance' in line.lower():
            results['ASA'] += 1

    # stores the key with the highest value in a variable
    firmware_version = max(results, key=results.get)
    firmware_object = {'IOS': IOS, 'IOSXE': IOSXE, 'NXOS': NXOS}.get(firmware_version)

    # returns firmware object
    return firmware_object






'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.TransportEngines import BaseEngine, default_buffer, default_timeout, \
    default_command_end, default_delay
from CiscoAutomationFramework.Exceptions import EnablePasswordError
from abc import ABC, abstractmethod
from inspect import getmodule


class CiscoFirmware(ABC):

    def __init__(self, transport):
        if not isinstance(transport, BaseEngine):
            raise TypeError(f'transport object MUST be an instance of {getmodule(BaseEngine).__name__}.{BaseEngine.__name__}')
        self._terminal_length_value = None
        self._terminal_width_value = None
        self.transport = transport
        #self.terminal_length()

    @property
    def is_nexus(self) -> bool:
        """
        Returns True/False if the Cisco device is a Nexus

        :return: True/False
        :rtype: bool
        """
        return False

    @property
    def commands_sent(self) -> list:
        """
        A list of all commands sent to the device in the order they are entered from first to last in the
        current session

        :return: List of commands sent
        :rtype: list
        """
        return self.transport.all_commands_sent


    def cli_to_config_mode(self) -> bool:
        """
        Navigates the CLI into config mode regardless of where it is

        :return: True/False
        :rtype: bool
        """
        if self.transport.in_user_exec_mode:
            self.cli_to_privileged_exec_mode()

        if self.transport.in_privileged_exec_mode:
            self.transport.send_command_get_output('config t')

        return self.transport.in_configuration_mode

    def cli_to_privileged_exec_mode(self) -> bool:  # TODO: Enable password typing will fail trying to get output. Fix!
        """
        Navigates the CLI into prvileged exec mode regardless of where it is. Will raise an EnablePasswordError
        if the transport engine does not have an enable password set and the network device asks for one.

        :return: True/False
        :rtype: bool

        :raises: CiscoAutomationFramework.Exceptions.EnablePasswordError
        """
        if self.transport.in_privileged_exec_mode:
            return True
        if self.transport.in_configuration_mode:
            self.transport.send_command_get_output('end')
            return self.transport.in_privileged_exec_mode

        if self.transport.in_user_exec_mode:
            enabling_output = self.transport.send_command_get_output('enable')
            if self.transport.prompt not in enabling_output:
                if not self.transport.enable_password:
                    raise EnablePasswordError('No enable password provided, network device is asking for one!')
                self.transport.send_command_get_output(self.transport.enable_password)
                return self.transport.in_privileged_exec_mode

    @property
    def prompt(self) -> str:
        """
        Returns the latest prompt that was returned from the device

        :return: Latest prompt returned from device
        :rtype: str
        """
        return self.transport.prompt

    @property
    def hostname(self) -> str:
        """
        Hostname of the Cisco device.

        :return: Hostname
        :rtype: str
        """
        return self.transport.hostname

    def send_command_get_output(self, command, end=default_command_end, buffer_size=default_buffer,
                                timeout=default_timeout, delay=default_delay) -> list:
        """
        Sends a command to the device and returns the output from the device. If there were multiple commands sent
        this will gather the output from all the commands at once. This command is a combination of send_command
        and get_output.

        :param command: Command to send
        :param end: End character (default \n)
        :param buffer_size: Size of buffer when getting output from device. You shouldnt have to modify this much
        :param timeout: Time to stop waiting for output if no output is received.
        :param delay: Time to wait before gathering output from device
        :return: Output from device starting with command, ending with prompt in a list split by line
        :rtype: list
        """
        return self.transport.send_command_get_output(command, end, buffer_size, timeout, delay)

    def send_command(self, command, end=default_command_end) -> None:
        """
        Sends command to device, does not get any output. You should not need to run this directly often

        :param command: Command to send
        :param end: End character (default \n)
        :return: Nothing
        """
        return self.transport.send_command(command, end)

    def get_output(self, buffer_size=default_buffer, timeout=default_timeout) -> list:
        """
        Gets output from the device until the prompt is returned or the timeout is reached. You should not need
        to run this directly often.

        :param buffer_size: Size of buffer when getting output from device. You shouldnt have to modify this much
        :param timeout: Time to stop waiting for output if no output is received.
        :return:
        """
        return self.transport.get_output(buffer_size, timeout)

    def send_question_get_output(self, command) -> list:
        """
        Special method to get the output from sending a command with a question mark. This will
        send the command with a trailing '?' get the output from that and then erase the command
        that is returned on the ending prompt. You should NOT put a question mark in the command
        this function will handle that for you but if you do, it will be removed so the CLI doesnt do something
        that is not expected

        :param command: Command to inspect
        :return: list of possible command completions, or next possible words in command
        :rtype: list
        """
        if command.endswith('?'):
            command = command.replace('?', '')
        question_output = self.send_command_get_output(command, end=' ?')

        # get rid of command that was sent from the terminal, to do that construct list of backspaces
        # The question mark is not returned as a part of the next prompt line but the space is, account for that.
        backspaces = ''.join([chr(8) for _char in range(len(command) + 1)])
        self.send_command_get_output(backspaces, end='')

        return question_output

    def close_connection(self) -> None:
        """
        Closes connection to device

        :return: Nothing
        """
        return self.transport.close_connection()

    def terminal_length(self, n='0'):
        """
        Sets terminal length of shell

        :param n: length
        :type n: str
        :return: Nothing
        """
        if self._terminal_length_value:
            if self._terminal_length_value != int(n):
                self._terminal_length_value = int(n)
                return self._terminal_length(n)
        else:
            self._terminal_length_value = int(n)
            return self._terminal_length(n)

    def terminal_width(self, n='0'):
        """
        Sets terminal width of shell

        :param n: width
        :type n: str
        :return:
        """
        if self._terminal_width_value:
            if self._terminal_width_value != int(n):
                self._terminal_width_value = int(n)
                return self._terminal_width(n)
        else:
            self._terminal_width_value = int(n)
            return self._terminal_width(n)

    # Begin abstract properties

    @property
    @abstractmethod
    def uptime(self) -> str:
        """
        Returns uptime of device

        :return: Uptime
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def interfaces(self) -> list:
        """
        List of interfaces on device

        :return: Interfaces
        :rtype: list
        """
        pass

    @property
    @abstractmethod
    def mac_address_table(self) -> str:
        """
        Returns the MAC address table in its raw form

        :return: MAC Address Table
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def arp_table(self) -> str:
        """
        Returns the arp table in its raw form

        :return: Arp Table
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def running_config(self) -> str:
        """
        Returns running config in its raw form

        :return: Running Configuration
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def startup_config(self) -> str:
        """
        Returns startup config in its raw form

        :return: Startup Config
        :rtype: str
        """
        pass

    # Begin abstract methods

    @abstractmethod
    def _terminal_length(self, n='0'):
        pass

    @abstractmethod
    def _terminal_width(self, n='0'):
        pass

    @abstractmethod
    def save_config(self):
        """
        Saves running config to startup config

        :return: Nothing
        """
        pass

    @abstractmethod
    def add_local_user(self, username, password, password_code=0, *args, **kwargs):
        """
        Method here should be generating a string that the network device accepts in the following format
        'username USERNAME <args> <kwarg key> <kwarg value>  password PASSWORD_CODE PASSWORD'

        :param username: Username to use
        :param password: Password to set (cleartext or encrypted depending on password_code)
        :param password_code: Code for password you are providing
        :param args: Single word arguments in command string
        :param kwargs: Key word arguments in command string
        :return: Nothing
        """
        pass

    @abstractmethod
    def delete_local_user(self, username):
        """
        Deletes a locally defined user on device

        :param username: username to delete
        :type username: str
        :return: Nothing
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.transport.close_connection()






class CDPEntry:

    def __init__(self, entry_data):
        self._data = entry_data

    @property
    def device_id(self):
        for line in self._data:
            if line.lower().startswith('device id'):
                return line.split(':')[1].strip()

    @property
    def ip_address(self):
        for line in self._data:
            if 'ip' in line.lower() and 'address' in line.lower():
                return line.split(':')[1].strip()

    @property
    def platform(self):
        for line in self._data:
            if line.lower().startswith('platform'):
                return line.split(':')[1].split(',')[0].strip()

    @property
    def local_interface(self):
        for line in self._data:
            if line.lower().startswith('interface') and 'address' not in line:
                return line.split(':')[1].split(',')[0].strip()

    @property
    def remote_interface(self):
        for line in self._data:
            if line.lower().startswith('interface') and 'address' not in line:
                return line.split(':')[-1].strip()

    @property
    def version(self):
        for index, line in enumerate(self._data):
            if line.lower().startswith('version:'):
                return self._data[index+1]

class CDPparser:

    def __init__(self, raw_cdp_data):
        if type(raw_cdp_data) is not list:
            raw_cdp_data.splitlines()
        self.raw_cdp_data = raw_cdp_data[2:-2]

    @property
    def _entries(self):
        all_sections = []
        section = []
        for line in self.raw_cdp_data:
            if line.startswith('---------'):
                if section:
                    all_sections.append(section)
                    section = []
                continue
            section.append(line)
        if section:
            all_sections.append(section)
        return all_sections

    @property
    def number_of_entries(self):
        return len(self.cdp_entries)

    @property
    def cdp_entries(self):
        return [CDPEntry(x) for x in self._entries]



interface_abbrivs = ('fa', 'gi', 'te', 'fo', 'eth')


class PowerInlineInterface:

    def __init__(self, interface_data):
        self._interface_data = interface_data

    @property
    def name(self):
        return self._interface_data.split()[0]

    @property
    def admin(self):
        return self._interface_data.split()[1]

    @property
    def oper(self):
        return self._interface_data.split()[2]

    @property
    def watts(self):
        return float(self._interface_data.split()[3])

    @property
    def detected_device(self):
        return ' '.join(self._interface_data.split()[4:-2])

    @property
    def poe_class(self):
        return self._interface_data.split()[-2]

    @property
    def max_watts(self):
        return float(self._interface_data.split()[-1])


class PowerInlineParser:

    def __init__(self, sh_power_inline):
        self._data = sh_power_inline

    @property
    def interfaces(self):
        return [PowerInlineInterface(line)
                for line in self._data if any([line.lower().startswith(x) for x in interface_abbrivs])]




'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.TransportEngines import SSHEngine
from CiscoAutomationFramework.FirmwareDetect import detect_firmware


def connect_ssh(ip, username, password, port=22, enable_password=None, timeout=10):
    """
    Connects to your cisco device, returns a firmware specific instance of CiscoFirmware object.

    :param ip: IP address or hostname of Cisco device
    :type ip: str

    :param username: Username used to login
    :type username: str

    :param password: Password for user
    :type password: str

    :param port:  Port to use (default 22)
    :type port: int

    :param enable_password: Enable password to use if the user does not have privileges directly to privilege exec
    :type enable_password: str

    :param timeout: SSH timeout in seconds
    :type timeout: int

    :return: CiscoFirmware Object
    :rtype: CiscoFirmware

    """

    engine = SSHEngine()
    engine.enable_password = enable_password
    engine.timeout = timeout
    engine.connect_to_server(ip, username, password, port)
    firmware = detect_firmware(engine)
    return firmware(engine)



from socket import inet_aton, error

__all__ = ['RoutingTableParser', 'Route', 'NexusRoute', 'IOSRoute']


def is_ip(string):
    # Must have 3 periods for ipv4 address
    num_periods = len([x for x in string if x == '.'])
    if num_periods != 3:
        return False
    # must be at least 7 characters (4 digits and 3 periods)
    if len(string) < 7:
        return False

    # rest of checking can be handled here
    try:
        inet_aton(string)
        return True
    except error:
        return False


class Route:
    def __init__(self, routing_table_entry):
        self.data = routing_table_entry


class NexusRoute(Route):

    @property
    def subnet(self):
        return self.data.split(',')[0].split('/')[0]

    @property
    def cidr(self):
        return self.data.split(',')[0].split('/')[1]

    @property
    def next_hops(self):
        ips = []
        for line in self.data.split('|')[1:]:
            ips.append(line.split()[1].replace(',', ''))

        return ips

    def tag(self, hop):
        for line in self.data.split('|'):
            if hop in line:
                for word in line.split(','):
                    if 'tag' in word.lower():
                        return line.split()[-1]
        return ''

    def preference(self, hop):
        for line in self.data.split('|'):
            if hop in line:
                for word in line.split(','):
                    if '[' in word:
                        return word.split('/')[0].replace('[', '').replace(']', '').strip()
        return ''

    def metric(self, hop):
        for line in self.data.split('|'):
            if hop in line:
                for word in line.split(','):
                    if '[' in word:
                        return word.split('/')[1].replace('[', '').replace(']', '').strip()
        return ''

    def __repr__(self):
        return f'{type(self).__name__}(Subnet: {self.subnet}/{self.cidr}, Num Next Hop:{len(self.next_hops)})'


class IOSRoute(Route):

    def _get_next_hops(self, flag_count=1):
        ips = []
        count = 0
        for line in self.data.split('|'):
            for word in line.split():
                word = word.replace(',', '')
                if is_ip(word):
                    count += 1
                if count > flag_count and is_ip(word):
                    ips.append(word)
        return ips


    @property
    def _subnet_with_cidr(self):
        for section in self.data.split():
            if is_ip(section.split('/')[0]):
                return section

    @property
    def subnet(self):
        return self._subnet_with_cidr.split('/')[0]

    @property
    def cidr(self):
        subnet_and_cidr = self._subnet_with_cidr.split('/')
        if len(subnet_and_cidr) < 2:
            return '32'
        else:
            return subnet_and_cidr[1]

    @property
    def next_hops(self):
        """
        Iterate over data, grabbing every IP address after the first (first one is the subnet) and excluding
        the address if 'subnetted' is in the same line of output which would not be a hop
        """
        ips = []
        count = 0
        for line in self.data.split('|'):
            for word in line.split():
                # remove any commas and if there is a CIDR notation, remove that
                word = word.replace(',', '')
                word = word.split('/')[0]
                if is_ip(word):
                    count += 1

                # if "subnetted" is in the line, it is notifying that the following output is a subnets of supernet
                # which means that the address in the line can be disregarded
                if count > 1 and is_ip(word) and 'subnetted' not in line:
                    ips.append(word)

        return ips

    def tag(self, hop):
        return ''

    def preference(self, hop):
        for line in self.data.split(','):
            if hop in line:
                return line.split('[')[1].split('/')[0]
        return ''

    def metric(self, hop):
        for line in self.data.split(','):
            if hop in line:
                return line.split(']')[0].split('/')[1].split('[')[-1]
        return ''

    def __repr__(self):
        return f'{type(self).__name__}(Subnet: {self.subnet}/{self.cidr})'


class RoutingTableParser:
    def __init__(self, raw_routing_table):
        self.unparsed = raw_routing_table

        if self.is_nexus:
            self._raw_data = self._parse_out_header_nexus(raw_routing_table)
        else:
            self._raw_data = self._parse_out_header_ios(raw_routing_table)

    @property
    def is_nexus(self):
        # if all the identifiers are somewhere in the first 5 lines of output, that is an IOS
        # routing table header
        identifiers = ['local', 'connected', 'static', 'rip', 'eigrp']
        if all(x in ' '.join(self.unparsed[:5]).lower() for x in identifiers):
            return False
        return True


    def _parse_out_header_ios(self, data):

        return_data = []
        flag = False
        for line in data:
            if 'gateway' in line.lower():
                flag = True
                continue
            if flag is True:
                return_data.append(line)
        return return_data

    def _parse_out_header_nexus(self, data):
        if 'ip route' in ' '.join(data[:3]).lower():
            data = data[3:]
        return_data = []
        flag = False
        for line in data:
            if line == '':
                flag = True
                continue
            if flag is True:
                return_data.append(line)

        return return_data

    def _str_startswith_char(self, string):
        if len(string) > 0:
            if string[0].isalnum():
                return True
        return False

    @property
    def structured_data(self):
        rightsized_lines = []
        tmp = ''
        for line in self._raw_data:
            if self._str_startswith_char(line):
                if tmp:
                    rightsized_lines.append(tmp)
                tmp = line
            else:
                if line:
                    tmp += f' |{line.strip()}'
        if tmp:
            rightsized_lines.append(tmp)
        return rightsized_lines

    @property
    def routes(self):
        if self.is_nexus:
            return [NexusRoute(x) for x in self.structured_data]
        else:
            return [IOSRoute(x) for x in self.structured_data]



from threading import Thread
from CiscoAutomationFramework import connect_ssh
from CiscoAutomationFramework.FirmwareBase import CiscoFirmware
from abc import ABC, abstractmethod


class SSH(Thread, ABC):

    """
    Base SSH object that you can inherit from when building a threaded script that will run across the network.

    To build a script you must subclass this object and override the methods that are documented.

    Method Execution Order
    The object logs into the device
    during_login is executed
    if perform_secondary_action is set to True secondary_action is executed and then post_secondary_action


    """

    def __init__(self, ip, username, password, enable_password=None, perform_secondary_action=False):
        """

        :param ip: IP address of device
        :type ip: str
        :param username: Username to login with
        :type username: str
        :param password: Password for user
        :type password: str
        :param enable_password: Enable password if login user does not have direct access to privilege exec mode
        :type enable_password: str
        :param perform_secondary_action: True/False to execute secondary_action method when logged in
        :type perform_secondary_action: bool
        """
        super().__init__()
        self.ip = ip
        self.username = username
        self.password = password
        self.enable_password = enable_password
        self.perform_secondary_action = perform_secondary_action
        self.hostname = ''
        self.commands_sent = []
        self.is_nexus = False

    def during_login(self, ssh):
        """
        Every time this object runs against a device this method will be run. Override this method and add logic
        that needs to be run every time your object runs against a device.

        I find I often add logic here that gathers data and does not modify configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    def secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True.

        I often find this is useful for containing logic that modifies configuration that only needs to be run if a
        certain condition is met.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    def post_secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True and AFTER secondary_action has completed.

        I often find this is useful for containing logic that re gathers data or saves configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    def run(self) -> None:

        with connect_ssh(self.ip, self.username, self.password) as ssh:
            self.is_nexus = ssh.is_nexus
            self.hostname = ssh.hostname
            self.during_login(ssh)
            if self.perform_secondary_action:
                self.secondary_action(ssh)
                self.post_secondary_action(ssh)
            self.commands_sent = ssh.commands_sent


class SSHSplitDeviceType(SSH):

    """
    Create a script by inheriting from this object if you are running the same script against both
    Nexus and IOS devices but the command are different on each platform or the sequence is different.
    Out of the box it will run the functions prepended with ios on ios and iosxe devices and functions
    prepended with nexus on nxos devices. Because it is implied that this script will be run against a mix
    of nexus and ios you must define the functions regardless if they will be used or not when inheriting from
    this object
    """

    @abstractmethod
    def ios_during_login(self, ssh):
        """
        Every time this object runs against a device that is detected to be NOT Nexus this method
        will be run. Override this method and add logic that needs to be run every time your object runs against a
        device.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type

        I find I often add logic here that gathers data and does not modify configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    @abstractmethod
    def nexus_during_login(self, ssh):
        """
        Every time this object runs against a device that is detected to be running Nexus this method
        will be run. Override this method and add logic that needs to be run every time your object runs against a
        device.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type

        I find I often add logic here that gathers data and does not modify configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    @abstractmethod
    def ios_secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True and the device is detected as not Nexus.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type

        I often find this is useful for containing logic that modifies configuration that only needs to be run if a
        certain condition is met.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    @abstractmethod
    def nexus_secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True and the device is detected Nexus.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type.

        I often find this is useful for containing logic that modifies configuration that only needs to be run if a
        certain condition is met.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    @abstractmethod
    def ios_post_secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True and AFTER secondary_action has completed
        and the device is detected to be NOT Nexus.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type.

        I often find this is useful for containing logic that re gathers data or saves configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    @abstractmethod
    def nexus_post_secondary_action(self, ssh):
        """
        This method will run only if perform_secondary_action is set to True and AFTER secondary_action has completed
        and the device is detected to be Nexus.

        Must be overridden when inheriting from this object regardless if your script will run against this device
        type.

        I often find this is useful for containing logic that re gathers data or saves configuration.

        :param ssh: SSH object is provided to this method for interacting with the end device
        :type ssh: CiscoFirmware
        :return: Nothing
        """
        pass

    def during_login(self, ssh):
        if self.is_nexus:
            self.nexus_during_login(ssh)
        else:
            self.ios_during_login(ssh)

    def secondary_action(self, ssh):
        if self.is_nexus:
            self.secondary_action(ssh)
        else:
            self.secondary_action(ssh)
            pass

    def post_secondary_action(self, ssh):
        if self.is_nexus:
            self.nexus_post_secondary_action(ssh)
        else:
            self.ios_post_secondary_action(ssh)


def start_threads(object, ips, username, password, enable_password=None,
                  perform_secondary_action=False, wait_for_threads=False):

    """
    This helper function is a quick and easy way to start your threads. Gives you an option for waiting for threads
    to complete also.
    It will return the thread objects from the function for further processing by the main thread.

    :param object: Object you have written to perform your task. Must be inherted from SSH or SSHSplitDeviceType
    :type object: [SSH, SSHSplitDeviceType]
    :param ips: List of IP addresses or hostnames to run your worker object against
    :type ips: list[str]
    :param username: Username to login with
    :type username: str
    :param password: Password for user
    :type password: str
    :param enable_password: Enable password if user does not have rights to privilege exec mode without enable password
    :type enable_password: str
    :param perform_secondary_action: True if you want to run the secondary_action method against device
    :type perform_secondary_action: bool
    :param wait_for_threads: Wait for all threads to complete before returning
    :type wait_for_threads: bool
    :return: List of threads either running or completed depending on if wait_for_threads is True/False
    :rtype: list[SSH, SSHSplitDeviceType, type(object)]
    """

    if not issubclass(object, SSH):
        raise TypeError('object MUST be a subclass of ThreadedSSH!')

    # Instantiate thread objects and start them
    threads = [object(ip=ip, username=username, password=password, enable_password=enable_password,
                      perform_secondary_action=perform_secondary_action) for ip in ips]
    for thread in threads:
        thread.start()

    # if user wants to wait for them, wait for threads to finish
    if wait_for_threads:
        for thread in threads:
            thread.join()

    return threads



'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.Parsers.ConfigParser import ConfigParser
from CiscoAutomationFramework.Parsers.RoutingTableParser import RoutingTableParser
from CiscoAutomationFramework.Parsers.CDPparser import CDPparser
from CiscoAutomationFramework.Parsers.PowerInlineParser import PowerInlineParser
from CiscoAutomationFramework.Parsers.MacAddressTableParser import MacAddressTableParser


'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.IOS import IOS


class IOSXE(IOS):
    """Because IOS and IOSXE are so similar from the CLI, I chose to inherit all methods and attributes
    from the IOS class. If it is determined that a method does not behave as expected it can be overrode
    here. This eliminates duplication of code for both firmware types."""
    pass



interface_abbrivs = ('fa', 'gi', 'te', 'fo', 'eth', 'po')


class MacAddressTableInterface:

    def __init__(self, name, mac_addresses):
        self._name = name
        self._mac_addresses = mac_addresses

    @property
    def name(self):
        return self._name

    @property
    def mac_addresses(self):
        return self._mac_addresses

    @property
    def num_of_macs(self):
        return len(self._mac_addresses)

    def has_mac(self, mac):
        return mac in self.mac_addresses

    def __repr__(self):
        return f'{type(self).__name__}({self.name}, num_of_macs: {self.num_of_macs})'


class MacAddressTableParser:

    def __init__(self, mac_address_table):
        self._table = mac_address_table

    @property
    def is_nexus(self):
        """Searches the entire mac address table, checking if any of a set of keywords
        are in it"""
        table_as_string = ' '.join(self._table).lower()
        key_words = ['legend', 'vpc', 'peer']
        if any([word in table_as_string for word in key_words]):
            return True
        return False

    def _extract_mac(self, line):
        """
        The MAC address on IOS is displayed in the 2nd column
         All    0100.0ccc.cccc    STATIC      CPU

        and on Nexus platform it is displayed in the 3rd column
        *    1     aaaa.bbbb.cccc   dynamic  0         F      F    Eth1/18

        so handle accordingly
        """
        if self.is_nexus:
            return line.split()[2]
        return line.split()[1]

    def _extract_interface(self, line):
        """
        The interface is displayed in the last column of output for both IOS and NXOS based on the
        systems I have used, No need to handle differently
        """
        return line.split()[-1]

    def _line_is_good_data(self, line):
        if self.is_nexus:
            if len(line.split()) == 8 and any(x in line.split()[-1].lower() for x in interface_abbrivs):
                return True
        else:
            if len(line.split()) == 4 and any(x in line.split()[-1].lower() for x in interface_abbrivs):
                return True
        return False

    def _parse(self):
        data = {}
        for line in self._table:
            # make sure we dont try to parse part of the header, a separator line, blank lines, or anything
            # that is not a mac address entry
            if not self._line_is_good_data(line):
                continue
            interface = self._extract_interface(line)
            mac = self._extract_mac(line)
            if any(interface.lower().startswith(x) for x in interface_abbrivs):
                if interface in data.keys():
                    data[interface].append(mac)
                else:
                    data[interface] = [mac]
        return data

    @property
    def table_entries(self):
        return [MacAddressTableInterface(interface, mac_addresses) for interface, mac_addresses in self._parse().items()]



'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from paramiko import SSHClient, AutoAddPolicy
from datetime import datetime, timedelta
from abc import ABC, abstractmethod
from time import sleep

default_command_end = '\n'
default_buffer = 100
default_timeout = 1
default_delay = .5
standard_prompt_endings = ('>', '#', '> ', '# ')


class BaseEngine(ABC):

    def __init__(self):
        self.hostname = ''
        self.prompt = None
        self.enable_password = None
        self.commands_sent_since_last_output_get = 0
        self.all_commands_sent = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connection()

    def _extract_prompt(self, output):
        """
        Attempts to get last line of output, if successful checks for the standard prompt ending in
        the last line, if one is found, it splits the line by the prompt ending and sets the prompt to the
        first element plus the prompt ending and quits checking if any remain
        """
        try:
            last_line_of_output = output[-1].strip()
        except IndexError:
            return
        for prompt_ending in standard_prompt_endings:
            if prompt_ending in last_line_of_output:
                self.prompt = f'{last_line_of_output.split(prompt_ending)[0]}{prompt_ending}'
                return

    def send_command(self, command, end=default_command_end):

        self.commands_sent_since_last_output_get += 1
        self.all_commands_sent.append(command)
        return self._send_command(command, end)

    def get_output(self, buffer_size=default_buffer, timeout=default_timeout, no_command_sent_previous=False):
        """
        Gets the output that is being returned from the network device. It checks how many commands were run
        since the last time it got output, and will get output until a prompt is encountered. Each command
        should return the prompt after execution so if the user enters 6 commands, this will loop 6 times
        to get the output of all 6 commands. It is up to the user to either get data between all 6 commands
        or parse it out theirselves. If no command is sent, to loop through a single time set the
        'no_command_sent_previous' flag to True so it auctually attempts to gather output. The reason
        for manually setting that flag is so that if the user attempts to get output when there is none
        and has a higher timeout it will not try and gather output and hang. If the user wants to get output
        without sending a command they should know to expect some output.

        This could run into issues if there is logging enabled on the terminal but that is not something
        that is being done inside the framework at this time so I dont believe there is a need to change that.
        """

        if no_command_sent_previous:
            self.commands_sent_since_last_output_get += 1

        output = ''
        for x in range(self.commands_sent_since_last_output_get):
            data = '\n'
            end = datetime.now() + timedelta(seconds=timeout)

            # while the last line of output DOES NOT START with the hostname AND DOES NOT CONTAIN a standard prompt ending
            while not all([data.splitlines()[-1].startswith(self.hostname), any([x in data.splitlines()[-1] for x in standard_prompt_endings])]):
                from_device = self._get_output(buffer_size)
                if from_device:
                    data += from_device
                    end = datetime.now() + timedelta(seconds=timeout)  # reset timeout clock
                else:
                    if datetime.now() > end:
                        # timeout clock triggered, break out of loop because we must be at a point
                        # in the CLI where it does not return a prompt or is hung
                        break
                    sleep(.1)
            output += data[1:]

        self.commands_sent_since_last_output_get = 0

        if type(output) != list:
            output = output.splitlines()

        self._extract_prompt(output)

        return output

    def send_command_get_output(self, command, end=default_command_end, buffer_size=default_buffer, timeout=default_timeout, delay=default_delay):
        self.send_command(command, end)
        if delay:
            sleep(delay)
        return self.get_output(buffer_size, timeout)

    @property
    def in_user_exec_mode(self) -> bool:
        if self.prompt.endswith('>'):
            return True
        return False

    @property
    def in_privileged_exec_mode(self) -> bool:
        if self.prompt.endswith('#') and not self.prompt.endswith(')#'):
            return True
        return False

    @property
    def in_configuration_mode(self) -> bool:
        if self.prompt.endswith(')#'):
            return True
        return False

    @abstractmethod
    def connect_to_server(self, ip, username, password, port) -> bool:
        pass

    @abstractmethod
    def _send_command(self, command, end) -> None:
        pass

    @abstractmethod
    def _get_output(self, buffer_size) -> str:
        pass

    @abstractmethod
    def close_connection(self) -> None:
        pass


class SSHEngine(BaseEngine):

    def __init__(self):
        super().__init__()
        self.client = SSHClient()
        self.client.set_missing_host_key_policy(AutoAddPolicy())
        self.shell = None
        self.timeout = 10

    def connect_to_server(self, ip, username, password, port):
        self.client.connect(hostname=ip, port=port, username=username, password=password, timeout=self.timeout)
        self.shell = self.client.invoke_shell()
        self.prompt, self.hostname = self._get_prompt_and_hostname()

    def _get_output(self, buffer_size):
        if self.shell.recv_ready():
            return bytes.decode(self.shell.recv(buffer_size))
        return ''

    def _send_command(self, command, end='\n'):
        self.shell.send(f'{command}{end}')

    def close_connection(self):
        self.client.close()

    # Methods required for some lower level SSH handling. These methods should not be called outside of this class

    def _send_space_get_data(self, timeout=1):
        self._send_command('', end='\n')
        end = datetime.now() + timedelta(seconds=timeout)
        data = ''
        while not data.endswith(standard_prompt_endings):
            from_device = self._get_output(1000)
            if from_device:
                data += from_device
                end = datetime.now() + timedelta(seconds=timeout)
            else:
                sleep(.1)
                if datetime.now() >= end:
                    break
        return data

    def _prompt_lookup(self, output_data):
        """
        Recursive method that will try and find the prompt in the last 3 lines of config, if not
        will send a space, get the data and call itself to run again
        """

        prompt = None
        for line in reversed(output_data.splitlines()[-3:]):
            if len(line.split()) == 1 and any(x in line for x in standard_prompt_endings):
                prompt = line.strip().replace('\r\n', '')
                break
        if not prompt:
            return self._prompt_lookup(self._send_space_get_data())

        return prompt

    def _get_prompt_and_hostname(self):
        # gets initial prompt out of device
        output = ''
        while not output.endswith(standard_prompt_endings):
            data = self._get_output(100)
            output += data
        prompt = self._prompt_lookup(output)

        # prompt = data.splitlines()[-1].strip()
        hostname = prompt[:-1]
        return prompt, hostname

    # End low level methods


'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.FirmwareBase import CiscoFirmware
from time import sleep

class NXOS(CiscoFirmware):

    @property
    def is_nexus(self):
        return True

    @property
    def uptime(self):
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')
        device_output = self.transport.send_command_get_output('show version')
        for line in device_output.splitlines():
            if 'uptime' in line.lower():
                return ' '.join(line.split()[3:])

        return None

    @property
    def interfaces(self):
        possible_interface_prefixes = ('Eth', 'Po', 'Vl', 'mgm ', 'Gi', 'Te', 'Fo', 'Fa')
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')
        raw_data = self.transport.send_command_get_output('show interface', buffer_size=500)
        try:
            parsed_data = [x.split()[0] for x in raw_data[2:-2] if len(x) > 1 and x.startswith(possible_interface_prefixes)]
        except IndexError as _:
            raise IndexError('Unexpected data from device, Unable to extract interface names from "show interfaces" command!')
        return parsed_data

    @property
    def mac_address_table(self):
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')
        raw_mac = self.transport.send_command_get_output('show mac address-table')
        return '\n'.join(raw_mac[8:-2])

    @property
    def arp_table(self):
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')
        raw_arp = self.transport.send_command_get_output('show ip arp')
        return '\n'.join(raw_arp[11:-2])

    @property
    def running_config(self):
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')
        self.transport.send_command('show running-config')

        running_config = self.transport.get_output(buffer_size=100)
        while len(running_config) < 4 and not any([True if self.prompt in x else False for x in reversed(running_config[-4:])]):
            running_config += self.transport.get_output(buffer_size=100, no_command_sent_previous=True)
            sleep(.1)
        return '\n'.join(running_config[2:-2])

    @property
    def startup_config(self):
        self.cli_to_privileged_exec_mode()
        self.terminal_length('0')

        config = self.transport.send_command_get_output('show startup-config', buffer_size=100)
        while len(config) < 4 and not any([True if self.prompt in x else False for x in reversed(config[-4:])]):
            config += self.transport.get_output(buffer_size=100, no_command_sent_previous=True)
            sleep(.1)
        return '\n'.join(config[2:-2])

    def _terminal_length(self, n='0'):
        self.cli_to_privileged_exec_mode()
        return self.transport.send_command_get_output(f'terminal length {n}')

    def _terminal_width(self, n='0'):
        self.cli_to_privileged_exec_mode()
        return self.transport.send_command_get_output(f'terminal width {n}')

    def save_config(self):
        self.cli_to_privileged_exec_mode()
        data = self.transport.send_command_get_output('copy running-config startup-config', timeout=15)
        # if the prompt is in the last line of output and there is not a percent sign in any line of output we will
        # interpret that as a succesful save
        if self.transport.prompt in ''.join(data[-1:]) and any('complete' in line for line in data):
            return True
        return False

    def add_local_user(self, username, password, password_code=0, *args, **kwargs):
        if len(password) < 8:
            raise Exception('Password must be at least 8 characters!')
        kwarg_string = ' '.join([f'{key} {value}' for key, value in kwargs.items()])

        # if there is a space in the password, wrap it in quotes
        if ' ' in password:
            password = f'"{password}"'

        command_string = f'username {username} {" ".join(args)} {kwarg_string} password {password_code} {password}'
        self.cli_to_config_mode()
        return self.transport.send_command_get_output(command_string)

    def delete_local_user(self, username):
        self.cli_to_config_mode()
        self.transport.send_command(f'no username {username}')
        return self.transport.send_command_get_output('')


#!/usr/bin/env python
from setuptools import setup, find_packages
#from CiscoAutomationFramework import __version__

__version__ = '1.0.0'

description = '''The CiscoAutomationFramework is designed to be an interface for network engineers to 
issue commands and retrieve output from Cisco devices regardless of firmware so they can easily build automation 
scripts atop the framework and dont have to worry about the nuances of CLI scraping.'''


setup(
    name='CiscoAutomationFramework',
    version=__version__,
    packages=find_packages(),
    keywords='cisco automation framework network',
    url='https://github.com/superadm1n/CiscoAutomationFramework',
    license='Apache 2.0',
    author='Kyle Kowalczyk',
    author_email='kowalkyl@gmail.com',
    description='Framework for issuing commands and retrieving consistent data on Cisco devices',
    long_description=description,
    install_requires=['paramiko', 'pyserial==3.4'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.9'
    ]
)



'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''


class ConfigSection:
    def __init__(self, raw_config):
        self.raw_config = raw_config

    def __eq__(self, other):
        if isinstance(other, type(self)) and self.raw_config == other.raw_config:
            return True
        return False


class InterfaceConfig(ConfigSection):

    @property
    def interface_name(self):
        return self.raw_config[0].split()[-1]

    def has_config(self, config_line):
        for line in self.raw_config:
            if config_line.lower() in line.lower():
                return True
        return False

    def __repr__(self):
        return f'{type(self).__name__}({self.interface_name})'


class IPAccessControlList(ConfigSection):

    @property
    def name(self):
        return self.raw_config[0].split()[-1]

    @property
    def type(self):
        return self.raw_config[0].split()[2]

    def __repr__(self):
        return f'{type(self).__name__}({self.name})'




def column_print(data, spaces_between_columns=2, separator_char=None):
    """
    Print in nice even columns where each column is only the width it needs to be, not the width of the
    widest column in the list.
    Based off of code from https://stackoverflow.com/questions/9989334/create-nice-column-output-in-python
    """
    # Get the max width of each column
    column_widths = []
    for index, _ in enumerate(data[0]):
        column_widths.append(max(len(str(row[index])) for row in data) + spaces_between_columns)

    # reconstruct list if a separator character is specified to be placed between the header and data
    if separator_char and type(separator_char) is str and len(separator_char) == 1:
        separator_line = [[separator_char * (x - spaces_between_columns) for x in column_widths]]
        first_line = data[:1]
        rest = data[1:]
        data = first_line + separator_line + rest


    # construct a string where each column is its own max width and print the line out when completed.
    for row in data:
        print_string = ''
        for index, word in enumerate(row):
            print_string += str(word).ljust(column_widths[index])
        print(print_string)


'''
Copyright 2021 Kyle Kowalczyk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
from CiscoAutomationFramework.Parsers.ConfigSectionTypes import InterfaceConfig, IPAccessControlList


class ConfigParser:

    def __init__(self, running_config):
        self.running_config = running_config.splitlines()

    def _nextline_startswith_space(self, current_line_index):
        try:
            if self.running_config[current_line_index + 1].startswith(' '):
                return True
        except IndexError:
            return False
        else:
            return False

    @property
    def _config_sections(self):
        """
        Parses running config for any first level configuration sections for example below would be a section

        router eigrp 1
          network 1.1.1.1 0.0.0.0
          network 2.2.2.2 0.0.0.0

        """
        data = []
        section_data = []
        for index, line in enumerate(self.running_config):
            if self._nextline_startswith_space(index):
                section_data.append(line)
            else:
                if len(section_data) > 0:
                    section_data.append(line)
                    data.append(section_data)
                    section_data = []
        return data

    def get_config_section(self, title_startswith, return_all=True):
        """
        Extracts a specific configuration section whos first line starts with your variable.
        Because we search if the title startswith your variable you could have multiple results
        so instead of always returning the first one, I give the option to return all

        """
        for section in self._config_sections:
            if section[0].startswith(title_startswith):
                if return_all:
                    yield section
                else:
                    return section

    def has_global_config(self, config_string):
        for line in self.running_config:
            if config_string in line and not line.startswith(' '):
                return True
        return False

    def interfaces_that_have_config(self, config_string):
        interfaces = []
        for interface in self.interface_configs:
            if interface.has_config(config_string):
                interfaces.append(interface)
        return interfaces

    @property
    def version(self):
        for line in self.running_config:
            if 'version' in line:
                return line.split()[1]
        return None

    @property
    def local_users(self):
        users = []
        for line in self.running_config:
            if line.lower().startswith('username'):
                users.append(line.split()[1])
        return users

    @property
    def interface_configs(self):
        return [InterfaceConfig(x) for x in self._config_sections if x[0].startswith('interface')]

    @property
    def ip_access_control_lists(self):
        return [IPAccessControlList(x) for x in self._config_sections if x[0].startswith('ip access-list')]

    @property
    def numbered_access_control_lists(self):
        data = []
        acl_data = []
        acl_num = None

        for line in self.running_config:
            if line.startswith('access-list'):
                num = line.split()[1]
                if num != acl_num:
                    if len(acl_data) > 0:
                        data.append(acl_data)
                        acl_data = []
                    acl_num = num
                    acl_data.append(line)

                else:
                    acl_data.append(line)
        if len(acl_data) > 0:
            data.append(acl_data)
        return data


